package require Tkogl
package require OpenGL
package require OpenGLU

#---------------------------------------------------------------------------
# 
#  The following are sample procs that compute specifications for the
#  gencyl command of the GL widget.
#
#---------------------------------------------------------------------------

proc cylinder { { radius 1 } { height 1 } { nsides 20 } } {
    # Returns a gencyl specification of a circular cylinder with
    # the given radius and height
    return "-close both -polygon $radius $nsides -plot -transl 0 0 $height -plot"
}

proc cone { { radius 1 } { height 1 } { nsections 10} { nsides 20 } } {
    # Returns a gencyl specification of a cone with circular base
    # with the given radius and height. 10 cross sections are
    # generated in order to yield a smoother shading
    set cyl "-polygon $radius $nsides "
    for {set i 0} {$i<=$nsections} {incr i} {
        lappend cyl -ident -transl 0 0 [expr $height*$i*1.0/$nsections]
        lappend cyl -scale [expr ($nsections-$i)*0.1] \
                [expr ($nsections-$i)*0.1] 1 -plot
    }
    return $cyl
}

proc gear { {inner 3} {outer 4} {width 1} {teeth 20} {ncross 4} {torsiondegrees 30} } {
   lappend cyl -close both -cross
   set teeth [expr 2*$teeth]
   for {set tooth 0} {$tooth<$teeth} {incr tooth} {
      set ang1 [expr 3.141592654*2*$tooth/$teeth]
      set ang2 [expr 3.141592654*2*($tooth+1)/$teeth]
      if {$tooth%2==0} { set radius $inner } { set radius $outer } 
      lappend cyl [expr $radius * cos($ang1)] [expr $radius * sin($ang1)] 0.0 
      lappend cyl [expr $radius * cos($ang2)] [expr $radius * sin($ang2)] 0.0 
   }
   lappend cyl -plot 
   set dtrans [expr -$width*1.0/$ncross]
   set drot [expr $torsiondegrees*1.0/$ncross]
   for {set i 1} { $i < $ncross } {incr i} {
       lappend cyl -translate 0 0 $dtrans -rotate $drot 0 0 1 -plot
   }
   return $cyl
}

proc torus { { inner 1 } { outer 10 } { nsections 36 } { nvertices 12 } } {
    # Returns the gencyl specification of a torus where the cross-section
    # has radius 'inner' and is rotated along the main axis at a distance
    # of 'radius'. 'nsections' specifies the number of cross sections
    # generated. 'nvertices' is the number of vertices per cross section
    lappend cyl -stitch both -polygon $inner $nvertices 
    set dang [expr 360.0 / $nsections]
    for { set i 0 } {$i < 360} {set i [expr $i + $dang]} {
        lappend cyl -ident -transl -$outer 0 0 -rot $i 0 1 0 -plot
    }
    return $cyl
}

proc sphere { { radius 1 } { nsections 18 } { nstripes 20 } } {
    # Returns the gencyl specification of a sphere with the given radius.
    # 'nsections' corresponds to the total number of cross-sections and
    # 'nstripes' corresponds to the number of vertices in each cross-section.
    set cyl "-polygon $radius $nstripes"
    set dang [expr 180.0 / $nsections]
    for {set i -90.0} {$i <= 90.0} {set i [expr $i+$dang]} {
        set ang [expr $i*3.141592654/180.0]
        set x [expr $radius * cos($ang)]
        set y [expr $radius * sin($ang)]
        lappend cyl -ident -transl 0 0 $y -scale $x $x 1 -plot
    }
    return $cyl
}

proc screw { {radius 1} {height 5} {nturns 1} {dsize 0.05} {dang 10} {sides 3} } {
    # Returns the gencyl specification of a screw, which is a polygon
    # successively translated scaled and rotated along the z axis.
    # 'radius' is the initial radius of the cross-section.
    # 'height' is the total height of the screw.
    # 'nturns' is the number of complete turns of the shape.
    # 'dsize' is how much the size of the shape is incremented at each step
    # 'dang' is how much the shape is turned at each step (degrees)
    # 'sides' is the number of sides in the cross-section.
    set cyl "-shade smoothrows -polygon $radius $sides -plot"
    if $sides==2 { lappend cyl -stitch none }
    set iz [expr $height*$dang/360.0/$nturns]
    set n [expr $nturns*360.0/$dang]
    for {set i 0} {$i < $n} {incr i} {
        set s [expr $i*$dsize+1]
        lappend cyl -ident -transl 0 0 [expr $iz*$i] \
                -rot [expr $i*$dang] 0 0 1 \
                -scale $s $s 1 -plot
    }
    return $cyl
}

proc snail { {r0 1} {d0 0} {alpha 2} {nturns 3} {dang 10} } {
    # r0 is the initial radius
    # d0 is the initial distance between the center of the cross-section and
    #    the axis
    # alpha is the ratio of size between two full turns of the cross-section
    # nturns is the total number of turns
    # dang is the angle increment between two cross sections (degrees)
    set cyl "-polygon $r0 20"
    set r $r0
    set finalang [expr $nturns * 360]
    for { set i 0 } {$i < $finalang} { incr i $dang } {
        set k [expr $i/360.0]
        set r [expr $r0*pow($alpha,$k)]
        set dst [expr 2*$r0*(1-pow($alpha,$k+1))/(1-$alpha)+$d0-$r]
        lappend cyl -ident -scale $r $r $r -transl $dst 0 0 \
                -rot -$i 0 1 0 -plot
    }
    return $cyl
}

proc funcrev { {xmin -2} {xmax 2} {nx 20} {nsections 10} 
               {yfunc {sin($x)+2}} } {
    # Plots a revolution of the curve yfunc(x) for xmin<=x<=xmax
    # sampled nx times in the x direction and with nsections sections.
    # yfunc should be a command that, given a value for x, computes
    # the corresponding y value
    set dx [expr ($xmax-$xmin)*1.0/($nx-1)]
    lappend cyl -stitch ends -cross
    for {set ix 0} {$ix<$nx} {incr ix} {
        set x [expr $xmin+$ix*$dx]
        set y [eval expr $yfunc]
        lappend cyl $x $y 0
    }
    set dang [expr 360.0/$nsections]
    for {set i 0} {$i<360} {set i [expr $i+$dang]} {
        lappend cyl -plot -rot $dang 1 0 0 
    }
    return $cyl
}

proc funcmesh {{xmin -5} {xmax 5} {ymin -5} {ymax 5} {nx 20} {ny 20} 
               {zfunc {2*cos(sqrt($x*$x+$y*$y))}} } {
    # Plots a mesh of zfunc (x,y) for points xmin<=x<=xmax and 
    # ymin<=y<=ymax sampled nx times in the x direction and ny times
    # in the z direction. zfunc should be a command that takes two
    # numeric variables (x,y) and computes the corresponding z value
    # as the result
    set dx [expr ($xmax-$xmin)*1.0/($nx-1)]
    set dy [expr ($ymax-$ymin)*1.0/($ny-1)]
    set cyl "-stitch none"
    for {set ix 0} {$ix<$nx} {incr ix} {
        set x [expr $xmin+$ix*$dx]
        lappend cyl -ident -transl 0 0 $x -cross
        for {set iy 0} {$iy<$ny} {incr iy} {
            set y [expr $ymin+$iy*$dy]
            set z [eval expr $zfunc]
            lappend cyl $y $z 0
        }
        lappend cyl -plot
    }
    return $cyl
}

#---------------------------------------------------------------------------
#
# General utility procs for the GL widget
#
#---------------------------------------------------------------------------

proc centercamera { min max varname } {
    # Computes a default camera that centers an object given its bounding box
    # 'min' and 'max' are lists of the form "x y z".
    # As side effects, elements 'center', 'eye', 'size', 'znear' and 'zfar'
    # of global array  'varname' are set
    upvar #0 $varname camera
    scan $min "%f %f %f" xmin ymin zmin
    scan $max "%f %f %f" xmax ymax zmax
    
    set camera(center) "[expr ($xmax+$xmin)/2] [expr ($ymax+$ymin)/2] \
            [expr ($zmax+$zmin)/2]"
    set total 0
    foreach c {x y z} {
        set side [expr ([set ${c}max]-[set ${c}min])/2]
        set total [expr $total+$side*$side]
    }
    set camera(size) [expr sqrt($total)]
    set camera(eye) "[expr ($xmax+$xmin)/2] [expr ($ymax+$ymin)/2] \
            [expr ($zmax+$zmin)/2+2*$camera(size)]"
    set camera(field) 60
    set camera(znear) $camera(size)
    set camera(zfar) [expr 3*$camera(size)]
}

proc redraw {} {
	# sets up widget 'gl1' for redraw. The object to be drawn should
	# have been set up previously and the corresponding display list should
	# be in global variable 'dl'. Other global variables are used
	# to describe the scene:
	# 'xrot' and 'yrot' are angles in degrees for rotating the object
	# 'pmode' should be a valid constant for option -polygonmode
	# 'camera' is a global array containing the various options for
	# the view command

	global xrot yrot dl pmode camera shade debug
	scan $camera(center) "%f %f %f" xc yc zc

	if {$shade=="flat"} {
		set shademodel flat
	} else {
		set shademodel smooth
	}

	glClear GL_DEPTH_BUFFER_BIT
	glClear GL_COLOR_BUFFER_BIT
	glPolygonMode GL_FRONT $pmode
	glShadeModel $shademodel
	glMatrixMode GL_PROJECTION
	glLoadIdentity
	gluPerspective 60 1 $camera(znear) $camera(zfar)
	glMatrixMode GL_MODELVIEW
	glLoadIdentity

	gluLookAt [lindex $camera(eye) 0] [lindex $camera(eye) 1] [lindex $camera(eye) 2] [lindex $camera(center) 0] [lindex $camera(center) 1] [lindex $camera(center) 2] 0 1 0
	glTranslatef $xc $yc $zc
	glRotatef $yrot 1 0 0
	glRotatef $xrot 0 1 0
	glTranslatef [expr -$xc] [expr -$yc] [expr -$zc]

	glCallList $dl

	.gl1 swapBuffer
}

proc makemodel args {
	# Calls the gencyl command of widget 'gl1' with the result of
	# the command line expressed in 'args'. Sets up a default camera
	# and deletes previous display lists that may have been defined earlier
    
	global dl camera shade adaptive adaptivethreshold debug
	
	if {$adaptive} {
		set adaptiveoption "-adaptive $adaptivethreshold"
	} else {
		set adaptiveoption ""
	}

	set info [eval .gl1 gencyl -texgen 0 1 0 1 $adaptiveoption -shade $shade [$args]]
	centercamera [getpropval $info min] [getpropval $info max] camera
	catch {.gl1 deletelist $dl}

	set dl [getpropval $info displaylist]
	redraw
}

proc createcheckerimg {} {
    # Creates a simple texture image called img
    image create photo tmp -width 2 -height 2
    image create photo img -width 64 -height 64
    tmp put {{#ffffff #ff0000} {#000000 #ffffff}} 
    img copy tmp -from 0 0 2 2 -zoom 4 4 -to 0 0 64 64
    rename tmp {}
}

#---------------------------------------------------------------------------
#
# Procs to set up a user interface to examine and modify the default
# arguments of a procedure.
# 
#---------------------------------------------------------------------------

proc updatedefarg { procname } {
    global body arglist defarg
    foreach arg $arglist {
        lappend argspec [list $arg $defarg($arg)]
    }
    eval [list proc $procname $argspec $body]
}

proc setupdefargs { path procname updatecmd } {
    frame $path 
    global arglist defarg body
    set arglist [info args $procname]
    set body [info body $procname]
    foreach arg $arglist {
        if ![info default $procname $arg tmp] {
            set defarg($arg) {}
        } else {
            set defarg($arg) $tmp
        }
        pack [frame $path.$arg] -side top -expand yes -fill x        
        label $path.$arg.lab -text $arg
        entry $path.$arg.ent -textvar defarg($arg) 
        bind $path.$arg.ent <Key-Return> "updatedefarg $procname; eval $updatecmd"
        pack $path.$arg.lab -side left
        pack $path.$arg.ent -side right
    }
    return $path
}

#---------------------------------------------------------------------------
#
# Main program
#
#---------------------------------------------------------------------------

source "setutils.tcl"
createcheckerimg

OGLwin .gl1 -height 400 -width 400 -refresh redraw
pack .gl1 -expand yes -fill both -side left
set xrot 0 
set yrot 0
set pmode fill

set doRotate 1

bind .gl1 <Button-1> { set x %x ; set y %y; set doRotate 0 }
bind .gl1 <ButtonRelease-1> { set doRotate 1 }
bind .gl1 <B1-Motion> { 
    incr xrot [expr %x - $x] 
    incr yrot [expr %y - $y] 
    set x %x
    set y %y
    redraw
}
bind .gl1 <Button-2> { 
    if {$pmode=="fill"} {set pmode line} {set pmode fill} 
    redraw
}
bind .gl1 <Enter> { focus .gl1 }

proc doStuff {} {
	global xrot yrot doRotate
	
	if {$doRotate} {
		incr xrot 2 
		incr yrot 2
		redraw
	}

	after 10 doStuff
}

.gl1 makeCurrent

glLightfv GL_LIGHT0 GL_AMBIENT [newfv4 1 1 1 1]
glLightfv GL_LIGHT0 GL_DIFFUSE [newfv4 1 1 1 1]
glLightfv GL_LIGHT0 GL_POSITION [newfv4 1 1 1 0]
glLightfv GL_LIGHT0 GL_SPECULAR [newfv4 1 1 1 1]
glEnable GL_LIGHTING
glEnable GL_LIGHT0
glEnable GL_DEPTH_TEST

.gl1 texImage2D 0 0 img

glEnable GL_DEPTH_TEST
glTexParameteri GL_TEXTURE_2D GL_TEXTURE_WRAP_S GL_REPEAT
glTexParameteri GL_TEXTURE_2D GL_TEXTURE_WRAP_T GL_REPEAT
glTexParameteri GL_TEXTURE_2D GL_TEXTURE_MAG_FILTER GL_NEAREST
glTexParameteri GL_TEXTURE_2D GL_TEXTURE_MIN_FILTER GL_NEAREST
glTexEnvi GL_TEXTURE_ENV GL_TEXTURE_ENV_MODE GL_MODULATE
glMaterialfv GL_FRONT GL_AMBIENT [newfv4 0.4 0.4 0.4 0]
glMaterialfv GL_FRONT GL_DIFFUSE [newfv4 0.5 0.5 0.5 0]
glMaterialfv GL_FRONT GL_SPECULAR [newfv4 0.5 0.5 0.5 0]
glMateriali GL_FRONT GL_SHININESS 50
glMaterialfv GL_BACK GL_AMBIENT [newfv4 0.5 0 0 0]
glMaterialfv GL_BACK GL_DIFFUSE [newfv4 0.5 0 0 0]
glMaterialfv GL_BACK GL_SPECULAR [newfv4 0 0 0 0]
glMateriali GL_BACK GL_SHIINESS 50

pack [frame .mainframe] -side left -fill y
foreach f {cmdframe shadeframe modelframe} {
    frame .$f -bd 2 -relief raised
    pack .$f -fill both -in .mainframe -expand yes
}


button .quit -text "Quit" -command exit
set editcmd {
    catch "destroy .editpopup.w"
    if $edit {
        if ![winfo exists .editpopup] { toplevel .editpopup }
        pack [setupdefargs .editpopup.w $mname "makemodel $mname"] -side right
    } else {
        catch "destroy .editpopup"
    }
}
checkbutton .edit -text "Edit" -command $editcmd
checkbutton .texture -text "texture" -var texture -command {
   if {$texture} {glEnable GL_TEXTURE_2D} {glDisable GL_TEXTURE_2D}
   .gl1 redraw
} 
pack .quit .edit .texture -in .cmdframe -side left -fill x

set shade smooth
set adaptive 0
set adaptivethreshold 30
set texture 0
radiobutton .flat -text "Flat Shading" -value flat -var shade \
        -command { makemodel $mname }
radiobutton .smooth -text "Smooth Shading" -value smooth -var shade \
        -command { makemodel $mname }
radiobutton .smoothrows -text "Smooth Shading (rows)" \
        -value smoothrows -var shade \
        -command { makemodel $mname }
radiobutton .smoothcols -text "Smooth Shading (columns)" \
        -value smoothcols -var shade \
        -command { makemodel $mname }
checkbutton .adaptive -text "Adaptive Shading" -var adaptive \
        -command { makemodel $mname }

entry .adaptivethreshold -textvar adaptivethreshold 
bind .adaptivethreshold <Key-Return> { makemodel $mname }
pack .flat .smooth .smoothrows .smoothcols \
        -in .shadeframe -expand yes -fill y -anchor w
pack .adaptive .adaptivethreshold \
        -in .shadeframe -side left -expand yes -fill y 

set mname cylinder
makemodel $mname
foreach model { cylinder sphere cone torus gear screw snail funcrev funcmesh } {
    radiobutton .modelframe.$model -text $model \
            -var mname -value $model \
            -command { makemodel $mname ; eval $editcmd }
    pack .modelframe.$model -expand yes -fill y -anchor w
}

doStuff
