package require Tkogl
package require OpenGL
package require OpenGLU

set xrot 0 
set yrot 0
set x 0
set y 0

#-----------------------------------------------------------
#
# The 5 following procs init the vertices and face lists 
# describing the 5 platonic polyhedra
#
#-----------------------------------------------------------

proc init_tetrahedron {} {
    global poly
    set a [expr 1.0/sqrt(3.0)]
    set poly(tetrahedron,vtx) [list \
	[list $a $a $a] [list $a -$a -$a] \
        [list -$a $a -$a] [list -$a -$a $a] ] 
    set poly(tetrahedron,face) {
        {0 3 1} {2 0 1} {3 0 2} {1 3 2}
    } 
}

proc init_hexahedron {} {
    global poly
    set a [expr 1.0/sqrt(3.0)]
    set poly(hexahedron,vtx) [list \
	[list $a $a $a] [list $a $a -$a] \
	[list -$a $a -$a] [list -$a $a $a] \
	[list -$a -$a $a] [list -$a -$a -$a] \
	[list $a -$a -$a] [list $a -$a $a] ]
    set poly(hexahedron,face) {
        {0 1 2 3} {3 2 5 4} {4 5 6 7}
        {7 6 1 0}  {1 6 5 2} {0 3 4 7}
    }
}

proc init_octahedron {} {
    global poly
    set poly(octahedron,vtx) {
	{1 0 0} {0 1 0} {-1 0 0} {0 -1 0} {0 0 1} {0 0 -1} 
    }
    set poly(octahedron,face) {
        {0 1 4} {1 2 4} {2 3 4} {3 0 4}
        {1 0 5} {2 1 5} {3 2 5} {0 3 5}
    } 
}

proc init_icosahedron {} {
    global poly
    set X 0.525731112119133606
    set Z 0.850650808352039932
    set poly(icosahedron,vtx)\
	[list [list  -$X 0.0 $Z] [list $X 0.0 $Z] \
	     [list -$X 0.0 -$Z] [list $X 0.0 -$Z] \
	     [list 0.0 $Z $X] [list 0.0 $Z -$X] \
	     [list 0.0 -$Z $X] [list 0.0 -$Z -$X] \
	     [list $Z $X 0.0] [list -$Z $X 0.0] \
	     [list $Z -$X 0.0] [list -$Z -$X 0.0] ]
    set poly(icosahedron,face) { 
	{4 0 1}    {9 0 4}    {5 9 4}    {5 4 8}
	{8 4 1}    {10 8 1}   {3 8 10}    {3 5 8}
	{2 5 3}    {7 2 3}    {10 7 3}    {6 7 10}
	{11 7 6}    {0 11 6}    {1 0 6}    {1 6 10}
	{0 9 11}    {11 9 2}    {2 9 5}    {2 7 11}
    }
}

proc init_dodecahedron {} {
    global poly
    set alpha [expr sqrt(2.0/(3.0 + sqrt(5.0)))/sqrt(3.0)]
    set beta [expr (1.0 + sqrt(6.0/(3.0 + sqrt(5.0)) - 2.0 + \
	        2.0*sqrt(2.0/(3.0 + sqrt(5.0)))))/sqrt(3.0)]
    set gamma [expr 1.0/sqrt(3.0)]
    set poly(dodecahedron,vtx) [list \
	[list -$alpha  0  $beta] \
	[list $alpha  0  $beta ] \
	[list -$gamma  -$gamma  -$gamma ] \
	[list -$gamma  -$gamma  $gamma ] \
	[list -$gamma  $gamma  -$gamma ] \
	[list -$gamma  $gamma  $gamma ] \
	[list $gamma  -$gamma  -$gamma ] \
	[list $gamma  -$gamma  $gamma ] \
	[list $gamma  $gamma  -$gamma ] \
	[list $gamma  $gamma  $gamma ] \
	[list $beta  $alpha  0 ] \
	[list $beta  -$alpha 0 ] \
	[list -$beta  $alpha  0 ] \
	[list -$beta  -$alpha 0 ] \
	[list -$alpha  0  -$beta ] \
	[list $alpha  0  -$beta ] \
	[list 0  $beta  $alpha ] \
	[list 0  $beta  -$alpha ] \
	[list 0  -$beta  $alpha ] \
	[list 0  -$beta  -$alpha]]
    set poly(dodecahedron,face) {
	{0 1 9 16 5} 	{1 0 3 18 7} 
	{1 7 11 10 9}	{11 7 18 19 6} 
	{8 17 16 9 10}	{2 14 15 6 19} 
	{2 13 12 4 14}	{2 19 18 3 13} 
	{3 0 5 12 13}   {6 15 8 10 11} 
	{4 17 8 15 14} 	{4 12 5 16 17} 
    } 
}

#-----------------------------------------------------------------
#
#  make_stellated creates a stellated polyhedron out of a regular
#  polyhedron by inserting a new vertex in the center of the faces
#  and joining it to the face vertices by triangles. ratio is the
#  relative distance between the inserted vertex and the origin: 
#  -1 brings it close to the origin, 0 corresponds to a vertex lying
#  on the respective face, while 1+ brings it farther away.
#
#----------------------------------------------------------------

proc make_stellated { pname newpname ratio } {
    global poly
    set nface [llength $poly($pname,face)]
    set nvtx [llength $poly($pname,vtx)]
    set poly($newpname,vtx) $poly($pname,vtx)
    for {set i 0} {$i<$nface} {incr i} {
	set v [lindex $poly($pname,face) $i]
        set nvtxface [llength $v]
        for {set k 0} {$k<3} {incr k} { set m($k) 0.0 }
        for {set j 0} {$j<$nvtxface} {incr j} {
            set vv [lindex $poly($pname,vtx) [lindex $v $j]]
            for {set k 0} {$k<3} {incr k} { 
		set m($k) [expr $m($k) + [lindex $vv $k]]
            }
        }
        for {set k 0} {$k<3} {incr k} {
            set m($k) [expr $m($k)/$nvtxface*(1.0+$ratio)]
        }
        lappend poly($newpname,vtx) [list $m(0) $m(1) $m(2)]
        for {set j 0} {$j<$nvtxface} {incr j} {
            set jj [expr ($j+1)%$nvtxface]
            lappend poly($newpname,face) [list \
                [lindex $v $j] [lindex $v $jj] $nvtx]
        }
        incr nvtx
    }
}

#----------------------------------------------------------------------
#
#  faces_of_vertex returns a list where each element is a list of the
#  form { f v0 v1 } where:
#     f is the number of a face of poly $pname adjacent to $ivtx
#     v0 is the circular predecessor of $ivtx
#     v1 is the circular successor of $ivtx
#
#----------------------------------------------------------------------

proc faces_of_vertex { pname ivtx } {
    global poly
    set iface 0
    foreach f $poly($pname,face) {
        set i [lsearch -exact $f $ivtx]
        if {$i!=-1} {
           set nv [llength $f]
           set i0 [expr ($i-1+$nv)%$nv]
           set i1 [expr ($i+1)%$nv]
           lappend tmp [list $iface [lindex $f $i0] [lindex $f $i1]]
        }
        incr iface
    }
    set head [lindex $tmp 0]
    set tmp [lrange $tmp 1 end]
    lappend elist $head
    while {$tmp!={}} {
        set edge [lindex $head 2]
        set i [lsearch -glob $tmp "* $edge *"]
        if $i==-1 {error "error ordering faces"}
        set head [lindex $tmp $i]
        set tmp [lreplace $tmp $i $i]
        lappend elist $head
    }
    return $elist
}

#----------------------------------------------------------------------
#
#  make_truncated makes a truncated polyhedron out of a regular 
#  polyhedron by creating a new face for every vertex of the original
#  polyhedron. ratio is the relative length of the edge which is
#  cut from each original edge.
#
#----------------------------------------------------------------------
proc make_truncated { pname newpname ratio } {
    global poly
    set nvtx [llength $poly($pname,vtx)]
    set ratio1 [expr 1.0-$ratio]
    set newnvtx 0
    for {set ivtx 0} {$ivtx<$nvtx} {incr ivtx} {
        set elist [faces_of_vertex $pname $ivtx]
        set prevnvtx [expr $newnvtx+[llength $elist]-1]
        set v0 [lindex $poly($pname,vtx) $ivtx]
        set newface {}
        foreach e $elist {
            set ivtx1 [lindex $e 2]
            set iface [lindex $e 0]
            set v1 [lindex $poly($pname,vtx) $ivtx1]
            set newv {}
            for {set c 0} {$c<3} {incr c} {             
                lappend newv [expr [lindex $v0 $c]*$ratio1+\
                                   [lindex $v1 $c]*$ratio]
            }
            lappend poly($newpname,vtx) $newv
            set newface [linsert $newface 0 $newnvtx]
            set vtxsubst($ivtx,$iface) [list $prevnvtx $newnvtx]
            set prevnvtx $newnvtx
            incr newnvtx
        }
        lappend poly($newpname,face) $newface
    }
    set iface 0
    foreach f $poly($pname,face) {
        set newface {}
        foreach ivtx $f {
            lappend newface $vtxsubst($ivtx,$iface)
        }
        lappend poly($newpname,face) [join $newface]
        incr iface
    }      
}

#----------------------------------------------------------------------
#
# plot_poly plots all faces of polyhedron $pname in display lists
# whose numbers are stored in global variable poly($pname,dlists) 
#
#---------------------------------------------------------------------- 
proc plot_poly { path pname } {
	global poly

	$path makeCurrent

	if [info exists poly($pname,dlists)] return
	set nface [llength $poly($pname,face)]

	for {set i 0} {$i < $nface} {incr i} {
		set v [lindex $poly($pname,face) $i]

		set dl [glGenLists 1]
		lappend poly($pname,dlists) $dl
		glNewList $dl GL_COMPILE

		set nvtxface [llength $v]
		for {set k 0} {$k < 3} {incr k} { set n($k) 0.0 }

		for {set j 0} {$j < $nvtxface} {incr j} {
			set vv [lindex $poly($pname,vtx) [lindex $v $j]]
			for {set k 0} {$k < 3} {incr k} { 
				set n($k) [expr $n($k) + [lindex $vv $k]]
			}
		}

		set nsize [expr sqrt($n(0)*$n(0)+$n(1)*$n(1)+$n(2)*$n(2))]

		glBegin GL_POLYGON
		glNormal3f [expr $n(0)/$nsize] [expr $n(1)/$nsize] [expr $n(2)/$nsize]

		for {set j 0} {$j < $nvtxface} {incr j} {
			set vv [lindex $poly($pname,vtx) [lindex $v $j]]
			glVertex3f [lindex $vv 0] [lindex $vv 1] [lindex $vv 2]
		}

		glEnd
		glEndList
	}
}

#----------------------------------------------------------------------
#
#  init_polybuild sets up the 'build' and 'require' data structures
#  which describe how a given polyhedron name is to be built
#
#----------------------------------------------------------------------

proc init_polybuild {} {
    global build require
    foreach p {tetrahedron hexahedron octahedron dodecahedron
               icosahedron} {
        set build($p) "init_$p"
        set require($p) $p
        set build([list stellated $p]) [list \
	    make_stellated $p [list stellated $p] 1] 
        set require([list stellated $p]) [list $p [list stellated $p]]
	set build([list sunken $p]) [list \
	    make_stellated $p [list sunken $p] -0.5]
	set require([list sunken $p]) [list $p [list sunken $p]]
        set build([list half-truncated $p]) [list \
	    make_truncated $p [list half-truncated $p] 0.333]
        set require([list half-truncated $p]) [list $p \
            [list half-truncated $p]]
        set build([list truncated $p]) [list \
	    make_truncated $p [list truncated $p] 0.5]
        set require([list truncated $p]) [list $p \
            [list truncated $p]]
    } 
}

#----------------------------------------------------------------------
#
#  polybuild builds a display list for a polyhedron given its name
#
#----------------------------------------------------------------------
proc polybuild { path type variation dlist mode } {
	global poly build require bname

	if {$variation=="regular"} {
		set bname $type
	} else {
		set bname [list $variation $type]
	}

	foreach p $require($bname) {
		if ![info exists poly($p)] {
			eval $build($p)
		}
	}

	plot_poly $path $bname
	
	$path makeCurrent

	glNewList $dlist GL_COMPILE

	switch $mode {
		GL_FILL -
		GL_LINE {
			glEnable GL_LIGHTING
			glDisable GL_STENCIL_TEST
			glPolygonMode GL_FRONT_AND_BACK $mode 		
			
			foreach dl $poly($bname,dlists) {
				glCallList $dl
			}
		}

		GL_LINE/GL_FILL {
			glEnable GL_STENCIL_TEST
			glClear GL_STENCIL_BUFFER_BIT
			glStencilFunc GL_ALWAYS 0 1
			glStencilOp GL_INVERT GL_INVERT GL_INVERT
			glDisable GL_LIGHTING
			glPolygonMode GL_FRONT_AND_BACK GL_LINE

			foreach dl $poly($bname,dlists) {
				glCallList $dl
				glStencilFunc GL_EQUAL 0 1
				glStencilOp GL_KEEP GL_KEEP GL_KEEP
				glPolygonMode GL_FRONT_AND_BACK GL_FILL
				glEnable GL_LIGHTING
				glCallList $dl
				glStencilFunc GL_ALWAYS 0 1
				glStencilOp GL_INVERT GL_INVERT GL_INVERT
				glPolygonMode GL_FRONT_AND_BACK GL_LINE
				glDisable GL_LIGHTING
				glCallList $dl
			}
		}
	}

	glEndList
}

proc redraw args {
    global mode type variation dlist
    polybuild .gl $type $variation $dlist $mode
    .gl redraw
}

proc dumppoly { filename } {
    global poly bname
    set fd [open $filename w]
    foreach f $poly($bname,face) {
	puts $fd [llength $f]
	foreach v $f {
	    puts $fd [lindex $poly($bname,vtx) $v]
	}
    }
    close $fd
}

init_polybuild
pack [OGLwin .gl -aspect 1 -stencil 8 -refresh draw] -side left -fill both -expand yes
set dlist [glGenLists 1]

pack [frame .cmd] -side left -fill y
pack [frame .cmd.main] -side top -fill x

# quit button
pack [button .cmd.main.quit -text "Quit" -command exit] -side left

# mode menu
set mode GL_FILL
tk_optionMenu .cmd.main.mode mode GL_FILL GL_LINE GL_LINE/GL_FILL
trace variable mode w redraw
pack .cmd.main.mode -side left

# polygon type box
pack [frame .cmd.type -relief sunken -borderwidth 3] -fill x 
set type hexahedron
foreach p {tetrahedron hexahedron octahedron dodecahedron
               icosahedron} {
    pack [radiobutton .cmd.type.$p -text $p \
	      -var type -value $p -command redraw] -anchor w
}

# polygon variation box
pack [frame .cmd.variation -relief sunken -borderwidth 3] -fill x 
set variation regular
foreach p {regular stellated sunken half-truncated truncated} {
    pack [radiobutton .cmd.variation.$p -text $p -var variation\
	     -value $p -command redraw] -anchor w
}

# setup of initial OpenGL display
polybuild .gl hexahedron regular $dlist $mode

.gl makeCurrent
glMatrixMode GL_PROJECTION
glLoadIdentity
gluPerspective 20 1 0.5 20
glMatrixMode GL_MODELVIEW
glLoadIdentity
gluLookAt 0 0 10 0 0 0 0 1 0
glPushMatrix
glMaterialfv GL_FRONT GL_AMBIENT [newfv4 0.3 0.3 0.3 0]
glMaterialfv GL_FRONT GL_DIFFUSE [newfv4 0.5 0.5 0.5 0]
glMaterialfv GL_FRONT GL_SPECULAR [newfv4 0.3 0.3 0.3 0]
glMateriali GL_FRONT GL_SHININESS 40
glLightfv GL_LIGHT0 GL_POSITION [newfv4 1 1 1 0]
glLightfv GL_LIGHT0 GL_AMBIENT [newfv4 1 1 1 0]
glLightfv GL_LIGHT0 GL_DIFFUSE [newfv4 1 1 1 0]
glLightfv GL_LIGHT0 GL_SPECULAR [newfv4 1 1 1 0]
glClearColor 0 0 1 0
glEnable GL_LIGHTING
glEnable GL_LIGHT0
glEnable GL_LIGHT1
glEnable GL_DEPTH_TEST

proc draw {} {
	global dlist

	glClear GL_COLOR_BUFFER_BIT
	glClear GL_DEPTH_BUFFER_BIT
	glClear GL_STENCIL_BUFFER_BIT

	glCallList $dlist

	.gl swapBuffer
}

proc doStuff {} {
	global xrot yrot doRotate
	
	if {$doRotate} {
		incr xrot 2
		incr yrot 2

		.gl makeCurrent
		glMatrixMode GL_MODELVIEW
		glPopMatrix
		glPushMatrix
		glRotatef $yrot 1 0 0
		glRotatef $xrot 0 1 0
		.gl redraw
	}

	after 10 doStuff
}

set doRotate 1

bind .gl <Button-1> { set x %x ; set y %y; set doRotate 0 }
bind .gl <ButtonRelease-1> { set doRotate 1 }

bind .gl <B1-Motion> { 
	incr xrot [expr \$x - %x] 
	incr yrot [expr %y - \$y] 
	set x %x
	set y %y

	.gl makeCurrent
	glMatrixMode GL_MODELVIEW
	glPopMatrix
	glPushMatrix
	glRotatef $yrot 1 0 0
	glRotatef $xrot 0 1 0
	.gl redraw
}

draw
doStuff
